/*
 * ASYN UInt32Digital driver for Altera PLL reconfiguration
 */
#include <epicsStdio.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <cantProceed.h>
#include <asynDriver.h>
#include <asynUInt32Digital.h>
#include <devLib.h>
#include <iocsh.h>

/*
 * ASYN MULTIDEVICE addresses
 */
#define PLL_COUNTER_M_PHASE_ADJUST_ADDRESS         0
#define PLL_COUNTER_C0_PHASE_ADJUST_ADDRESS        1
#define PLL_COUNTER_C1_PHASE_ADJUST_ADDRESS        2
#define PLL_COUNTER_C2_PHASE_ADJUST_ADDRESS        3
#define PLL_COUNTER_C3_PHASE_ADJUST_ADDRESS        4
#define PLL_COUNTER_C4_PHASE_ADJUST_ADDRESS        5
#define PLL_COUNTER_C5_PHASE_ADJUST_ADDRESS        6
#define PLL_COUNTER_M_DIVISOR_ADDRESS              7
#define PLL_COUNTER_N_DIVISOR_ADDRESS              8
#define PLL_COUNTER_C0_DIVISOR_ADDRESS             9
#define PLL_COUNTER_C1_DIVISOR_ADDRESS            10
#define PLL_COUNTER_C2_DIVISOR_ADDRESS            11
#define PLL_COUNTER_C3_DIVISOR_ADDRESS            12
#define PLL_COUNTER_C4_DIVISOR_ADDRESS            13
#define PLL_COUNTER_C5_DIVISOR_ADDRESS            14
#define PLL_COUNTER_C5_DIVISOR_ADDRESS            14
#define PLL_SPREAD_BYPASS_ADDRESS                 15
#define PLL_CHARGE_PUMP_CURRENT_ADDRESS           16
#define PLL_LOOP_FILTER_RESISTOR_ADDRESS          17
#define PLL_LOOP_FILTER_CAPACITOR_ADDRESS         18
#define PLL_POR_RESET_STATUS_ADDRESS              19
#define PLL_FORCE_PLL_RESET_ADDRESS               20

/*
 * CSR values defined by Altera PLL reconfiguration component
 */
#define PLL_CSR_COUNTER_N               0x0
#define PLL_CSR_COUNTER_M               0x1
#define PLL_CSR_COUNTER_CP_LF           0x2
#define PLL_CSR_COUNTER_C0              0x4
#define PLL_CSR_COUNTER_C1              0x5
#define PLL_CSR_COUNTER_C2              0x6
#define PLL_CSR_COUNTER_C3              0x7
#define PLL_CSR_COUNTER_C4              0x8
#define PLL_CSR_COUNTER_C5              0x9
#define PLL_CSR_COUNTER_C(i)            (0x4+(i))

/*
 * Data port values defined by Altera PLL reconfiguration component
 */
#define PLL_PARAM_NM_NOMINAL_COUNT      0x00
#define PLL_PARAM_Cx_HI_CYCLES          0x00
#define PLL_PARAM_Cx_LO_CYCLES          0x10
#define PLL_PARAM_PHASE_STEP            0x20
#define PLL_PARAM_COUNTER_BYPASS        0x40
#define PLL_PARAM_COUNTER_Cx_ODD        0x50
#define PLL_PARAM_SPREAD_COUNTER_BYPASS 0x50
#define PLL_PARAM_CP_CURRENT            0x00
#define PLL_PARAM_LF_RESISTOR           0x10
#define PLL_PARAM_LF_CAPACITOR          0x20

/*
 * CSR values defined by PllReconfigIO logic
 */
#define PLL_CSR_DATA_MASK       0x01FF
#define PLL_CSR_CMD_LATCH_DATA  0x0200
#define PLL_CSR_CMD_RECONFIGURE 0x0400
#define PLL_CSR_CMD_READ        0x0800
#define PLL_CSR_CMD_WRITE       0x1000
#define PLL_CSR_CMD_RESET       0x2000
#define PLL_CSR_CMD_CLEAR_POR   0x4000
#define PLL_CSR_POR             0x4000
#define PLL_CSR_BUSY            0x8000

/*
 * Driver private storage
 */
typedef struct dpvt {
    epicsUInt16       *base;
    asynInterface      asynCommon;
    asynInterface      asynUInt32Digital;
} dpvt;

/*
 * Try to work around PLL reconfiguration components that never come ready
 */
static int alterPllReconfigWorkaround = 1;

/*
 * Hardware access
 */
static int
pllRead(dpvt *pdpvt, asynUser *pasynUser, int param, int counter)
{
    volatile epicsUInt16 *csrp = pdpvt->base;
    int i = 0;

    *csrp = PLL_CSR_CMD_READ | param | counter;
    while((*csrp & PLL_CSR_BUSY) != 0) {
        if (++i == 1000) {
            if (alterPllReconfigWorkaround) {
                printf("Warning -- PLL reconfiguration circuit never ready.\n");
                break;
            }
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                  "PLL reconfiguration circuit never ready");
            return -1;
        }
    }
    return *csrp & PLL_CSR_DATA_MASK;
}

static int
pllWrite(dpvt *pdpvt, asynUser *pasynUser, int cmd, int v)
{
    volatile epicsUInt16 *csrp = pdpvt->base;
    int i = 0;

    *csrp = PLL_CSR_CMD_LATCH_DATA | (v & PLL_CSR_DATA_MASK);
    *csrp = PLL_CSR_CMD_WRITE | cmd;
    i = 0;
    while ((*csrp & PLL_CSR_BUSY) != 0) {
        if (++i == 1000) {
            if (alterPllReconfigWorkaround) {
                printf("Warning -- PLL reconfiguration circuit never ready.\n");
                break;
            }
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                  "PLL reconfiguration circuit never ready");
            return -1;
        }
    }
    return 0;
}

static int
pllReconfig(dpvt *pdpvt, asynUser *pasynUser)
{
    volatile epicsUInt16 *csrp = pdpvt->base;
    int i = 0;

    *csrp = PLL_CSR_CMD_RECONFIGURE;
    i = 0;
    while ((*csrp & PLL_CSR_BUSY) != 0) {
        if (++i == 1000) {
            if (alterPllReconfigWorkaround) {
                printf("Warning -- PLL reconfiguration circuit never ready.\n");
                break;
            }
            epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                                  "PLL reconfiguration circuit never ready.");
            return -1;
        }
    }
    return 0;
}

/*
 * asynCommon methods
 */
static void
showCount(dpvt *pdpvt, asynUser *pasynUser, FILE *fp, int cIndex)
{
    int c = PLL_CSR_COUNTER_C(cIndex);
    int bypass;
    int low;
    int high;
    int odd;

    if (((bypass = pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_BYPASS, c)) < 0)
     || ((low = pllRead(pdpvt, pasynUser, PLL_PARAM_Cx_LO_CYCLES, c)) < 0)
     || ((high = pllRead(pdpvt, pasynUser, PLL_PARAM_Cx_HI_CYCLES, c)) < 0)
     || ((odd = pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_Cx_ODD, c)) < 0)) {
        fprintf(fp, "Failed to read counter code %d\n", c);
        return;
    }
    if (bypass) {
        low = 0;
        high = 1;
        odd = 1;
    }
    fprintf(fp, "    C%d high:low cycles count: %d%s:%d%s", cIndex, 
                                odd ? high - 1 : high, odd ? ".5" : "",
                                low,                   odd ? ".5" : "");
    fprintf(fp, "   Frequency = VCO/%d\n", high + low);
}

static void
pllReconfigReport(void *pvt, FILE *fp, int details)
{
    dpvt *pdpvt = (dpvt *)pvt;
    asynUser *pasynUser;
    int i;

    if (details <= 0)
        return;
    pasynUser = pasynManager->createAsynUser(NULL, NULL);
    fprintf(fp, "    PLL VCO Frequency = Input Clock Frequency");
    if (pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_BYPASS,
                                    PLL_CSR_COUNTER_M) == 0)
        fprintf(fp, " * %d",pllRead(pdpvt, pasynUser, PLL_PARAM_NM_NOMINAL_COUNT,
                                        PLL_CSR_COUNTER_M));
    if (pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_BYPASS,
                                    PLL_CSR_COUNTER_N) == 0)
        fprintf(fp, " / %d",pllRead(pdpvt, pasynUser, PLL_PARAM_NM_NOMINAL_COUNT,
                                        PLL_CSR_COUNTER_N));
    fprintf(fp, "\n");
    if (details <= 1)
        return;
    fprintf(fp, "    N spread counter bypass bit: %d\n",
        pllRead(pdpvt, pasynUser, PLL_PARAM_SPREAD_COUNTER_BYPASS, PLL_CSR_COUNTER_N));
    fprintf(fp, "    M spread counter bypass bit: %d\n",
        pllRead(pdpvt, pasynUser, PLL_PARAM_SPREAD_COUNTER_BYPASS, PLL_CSR_COUNTER_M));
    fprintf(fp, "    Charge pump current: %d\n",
            pllRead(pdpvt, pasynUser, PLL_PARAM_CP_CURRENT, PLL_CSR_COUNTER_CP_LF));
    fprintf(fp, "    Loop filter resistor: %d\n",
            pllRead(pdpvt, pasynUser, PLL_PARAM_LF_RESISTOR, PLL_CSR_COUNTER_CP_LF));
    fprintf(fp, "    Loop filter capacitor: %d\n",
            pllRead(pdpvt, pasynUser, PLL_PARAM_LF_CAPACITOR, PLL_CSR_COUNTER_CP_LF));
    for (i = 0 ; i < 6 ; i++)
        showCount(pdpvt, pasynUser, fp, i);
    pasynManager->freeAsynUser(pasynUser);
}

static asynStatus
pllReconfigConnect(void *pvt, asynUser *pasynUser)
{
    pasynManager->exceptionConnect(pasynUser);
    return asynSuccess;
}

static asynStatus
pllReconfigDisconnect(void *pvt, asynUser *pasynUser)
{
    pasynManager->exceptionDisconnect(pasynUser);
    return asynSuccess;
}

static asynCommon asynCommonMethods = { pllReconfigReport,
                                        pllReconfigConnect,
                                        pllReconfigDisconnect };

/*
 * asynUInt32Digital methods
 */
static asynStatus
pllReconfigRead(void *pvt, asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask)
{
    dpvt *pdpvt = (dpvt *)pvt;
    asynStatus status;
    int address;
    int low, high;
    int i;
    volatile epicsUInt16 *csrp = pdpvt->base;
    int c = 0;
    int p = 0;

    if ((status = pasynManager->getAddr(pasynUser, &address)) != asynSuccess)
        return status;
    switch (address) {
    case PLL_COUNTER_M_DIVISOR_ADDRESS:  c = PLL_CSR_COUNTER_M;  break;
    case PLL_COUNTER_N_DIVISOR_ADDRESS:  c = PLL_CSR_COUNTER_N;  break;
    case PLL_COUNTER_C0_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C0; break;
    case PLL_COUNTER_C1_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C1; break;
    case PLL_COUNTER_C2_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C2; break;
    case PLL_COUNTER_C3_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C3; break;
    case PLL_COUNTER_C4_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C4; break;
    case PLL_COUNTER_C5_DIVISOR_ADDRESS: c = PLL_CSR_COUNTER_C5; break;

    case PLL_SPREAD_BYPASS_ADDRESS: p = PLL_PARAM_SPREAD_COUNTER_BYPASS; break;
    case PLL_CHARGE_PUMP_CURRENT_ADDRESS:   p = PLL_PARAM_CP_CURRENT;   break;
    case PLL_LOOP_FILTER_RESISTOR_ADDRESS:  p = PLL_PARAM_LF_RESISTOR;  break;
    case PLL_LOOP_FILTER_CAPACITOR_ADDRESS: p = PLL_PARAM_LF_CAPACITOR; break;
    }
    switch (address) {
    case PLL_COUNTER_M_DIVISOR_ADDRESS:
    case PLL_COUNTER_N_DIVISOR_ADDRESS:
        i = pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_BYPASS, c);
        if (i < 0) {
            return asynError;
        }
        else if (i == 0) {
            i = pllRead(pdpvt, pasynUser, PLL_PARAM_NM_NOMINAL_COUNT, c);
            if (i < 0)
                return asynError;
            *value = i;
        }
        else {
            *value = 1;
        }
        break;

    case PLL_COUNTER_C0_DIVISOR_ADDRESS:
    case PLL_COUNTER_C1_DIVISOR_ADDRESS:
    case PLL_COUNTER_C2_DIVISOR_ADDRESS:
    case PLL_COUNTER_C3_DIVISOR_ADDRESS:
    case PLL_COUNTER_C4_DIVISOR_ADDRESS:
    case PLL_COUNTER_C5_DIVISOR_ADDRESS:
        i = pllRead(pdpvt, pasynUser, PLL_PARAM_COUNTER_BYPASS, c);
        if (i < 0) {
            return asynError;
        }
        else if (i == 0) {
            low = pllRead(pdpvt, pasynUser, PLL_PARAM_Cx_LO_CYCLES, c);
            if (low < 0)
                return asynError;
            high = pllRead(pdpvt, pasynUser, PLL_PARAM_Cx_HI_CYCLES, c);
            if (high < 0)
                return asynError;
            *value = low + high;
        }
        else {
            *value = 1;
        }
        break;

    case PLL_SPREAD_BYPASS_ADDRESS:
    case PLL_CHARGE_PUMP_CURRENT_ADDRESS:
    case PLL_LOOP_FILTER_RESISTOR_ADDRESS:
    case PLL_LOOP_FILTER_CAPACITOR_ADDRESS:
        i = pllRead(pdpvt, pasynUser, p, PLL_CSR_COUNTER_CP_LF);
        if (i < 0)
            return asynError;
        *value = i;
        break;

    case PLL_POR_RESET_STATUS_ADDRESS:
        *value = (*csrp & PLL_CSR_POR) != 0;
        break;

    default:
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                            "Invalid uint32Digital read address %d", address);
        return asynError;
    }
    return asynSuccess;
}

static asynStatus
pllReconfigWrite(void *pvt, asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
    dpvt *pdpvt = (dpvt *)pvt;
    asynStatus status;
    int address;
    int c;
    volatile epicsUInt16 *csrp = pdpvt->base;

    if ((status = pasynManager->getAddr(pasynUser, &address)) != asynSuccess)
        return status;
    switch (address) {
    case PLL_COUNTER_M_PHASE_ADJUST_ADDRESS:  c = PLL_CSR_COUNTER_M;  break;
    case PLL_COUNTER_C0_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C0; break;
    case PLL_COUNTER_C1_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C1; break;
    case PLL_COUNTER_C2_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C2; break;
    case PLL_COUNTER_C3_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C3; break;
    case PLL_COUNTER_C4_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C4; break;
    case PLL_COUNTER_C5_PHASE_ADJUST_ADDRESS: c = PLL_CSR_COUNTER_C5; break;

    case PLL_POR_RESET_STATUS_ADDRESS:
        if (value)
            *csrp = PLL_CSR_CMD_CLEAR_POR;
        return asynSuccess;

    case PLL_FORCE_PLL_RESET_ADDRESS:
        if (value) {
            volatile int i;
            *csrp = PLL_CSR_CMD_RESET;
            for (i = 0 ; i < 20 ; i++) continue;
            *csrp = 0;
        }
        return asynSuccess;

    default:
        epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
                            "Invalid uint32Digital write address %d", address);
        return asynError;
    }
    if (value != 0) {
        int i, n, v;

        if ((epicsInt32)value > 0) {
            v = 1;
            n = value;
        }
        else {
            v = 3;
            n = -value;
        }
        if (n > 50) {
            printf("Unreasonable PLL phase change request!\n");
            return asynError;
        }
        if (pllWrite(pdpvt, pasynUser, PLL_PARAM_PHASE_STEP | c, v) < 0)
            return asynError;
        for (i = 0 ; i < n ; i++)
            if (pllReconfig(pdpvt, pasynUser) < 0)
                return asynError;
        if (pllWrite(pdpvt, pasynUser, PLL_PARAM_PHASE_STEP | c, 0) < 0)
            return asynError;
    }
    return asynSuccess;
}

static asynUInt32Digital asynUInt32DigitalMethods = { pllReconfigWrite,
                                                      pllReconfigRead };

/*
 * Configure and register a PLL reconfiguration instance
 */
static void
pllReconfigConfigure(const char *portName,
                     const char *fpgaType,
                     int vmeAddress,
                     const char *vmeAddressSpace)
{
    dpvt *pdpvt;
    volatile void *base;
    size_t size = sizeof(epicsUInt16);
    char c;
    int addressSpace;
    epicsUInt16 dummy;
    asynStatus status;

    if ((epicsStrCaseCmp(fpgaType, "Stratix II") != 0)
     &&  (epicsStrCaseCmp(fpgaType, "StratixII") != 0)) {
        printf("Unknown FPGA type\n");
        return;
    }
    if (vmeAddressSpace == NULL)
        vmeAddressSpace = "A16";
    if ((sscanf(vmeAddressSpace, "%c%d", &c, &addressSpace) != 2)
     || (c != 'A')) {
        printf("Bad VME address space argument\n");
        return;
    }
    switch (addressSpace) {
    case 16: addressSpace = atVMEA16; break;
    case 24: addressSpace = atVMEA24; break;
    case 32: addressSpace = atVMEA32; break;
    default:
        printf("Unknown VME address space (A16, A24, A32 supported)\n");
        return;
    }
    pdpvt = callocMustSucceed(sizeof(dpvt), 1, portName);
    if (devRegisterAddress(portName, addressSpace, vmeAddress, size, &base) != 0) {
        printf("Can't register VME address %#x (%#x bytes).\n", vmeAddress, (unsigned int)size);
        return;
    }
    pdpvt->base = (epicsUInt16 *)base;
    if (devReadProbe(sizeof(dummy), base, &dummy) != 0) {
        printf("Memory probe failed\n");
        return;
    }

    /*
     * Register with ASYN
     */
    status = pasynManager->registerPort(portName, ASYN_MULTIDEVICE, 1, 0, 0);
    if(status!=asynSuccess) {
        printf("registerDriver failed\n");
        return;
    }
    pdpvt->asynCommon.interfaceType = asynCommonType;
    pdpvt->asynCommon.pinterface  = (void *)&asynCommonMethods;
    pdpvt->asynCommon.drvPvt = pdpvt;
    status = pasynManager->registerInterface(portName, &pdpvt->asynCommon);
    if (status!=asynSuccess){
        printf("registerInterface failed\n");
        return;
    }
    pdpvt->asynUInt32Digital.interfaceType = asynUInt32DigitalType;
    pdpvt->asynUInt32Digital.pinterface  = (void *)&asynUInt32DigitalMethods;
    pdpvt->asynUInt32Digital.drvPvt = pdpvt;
    status = pasynManager->registerInterface(portName, &pdpvt->asynUInt32Digital);
    if (status!=asynSuccess){
        printf("registerInterface failed\n");
        return;
    }
}

/*
 * IOC shell command registration
 */
static const iocshArg pllReconfigConfigureArg0 = { "name",iocshArgString};
static const iocshArg pllReconfigConfigureArg1 = { "FPGA type",iocshArgString};
static const iocshArg pllReconfigConfigureArg2 = { "VME address",iocshArgInt};
static const iocshArg pllReconfigConfigureArg3 = { "[Axx]",iocshArgString};
static const iocshArg *pllReconfigConfigureArgs[] = {
    &pllReconfigConfigureArg0, &pllReconfigConfigureArg1,
    &pllReconfigConfigureArg2, &pllReconfigConfigureArg3};
static const iocshFuncDef pllReconfigConfigureFuncDef =
              {"alteraPllReconfigConfigure",4,pllReconfigConfigureArgs};
static void pllReconfigConfigureCallFunc(const iocshArgBuf *args)
{
    pllReconfigConfigure(args[0].sval, args[1].sval, args[2].ival, args[3].sval);
}
void
alteraPllReconfigRegistrar(void)
{
    iocshRegister(&pllReconfigConfigureFuncDef,pllReconfigConfigureCallFunc);
}
epicsExportRegistrar(alteraPllReconfigRegistrar);
